#include "simulation.h"
#include "random.h"
#include <time.h>
#include <iostream>
//-------------------------------SHO demo--------------------------------

class TRequest : public CProcess                 // request in the queueing system
{
	virtual void Run();                            // request's behaviour
public:
	int PortIndex;
	int VelikostRamce;

};

class TService : public CProcess                 // process servicing requests
{
	virtual void Run();                            // service's behaviour
public :
	int Id;
};


const PortCount = 4;
TService  *Service[PortCount];
CHead *Queue[PortCount];
int QBytes[4];

double Speed = 0.012500; // B/ns
int PortTable[PortCount][PortCount] = { // % [in, out]
		0, 20, 50, 30,
//		0, 0, 0, 100,
		20, 0, 40, 40,
		45, 30, 0, 25,
		30, 40, 30, 0
		};
const MeziramcovaMezera = 960; // ns
double PortRamce[PortCount] = { // ns
		(double)1 / 5000000,
		(double)1 / 20000000,
		(double)1 / 12000000,
		(double)1 / 8000000
	};

// Stat
double MaximalniDoba;
double CelkovaDoba;
int PocetRamcu;

double MaximalniDobaP[PortCount];
double CelkovaDobaP[PortCount];
int PocetRamcuP[PortCount];
int PocetRamcuP2[PortCount];

int MaxDelkaFront[PortCount];
int MaxDelkaFrontB[PortCount];

void TRequest::Run()
{
	double IncomingTime = Time();
	printf("Prisel ramec v case %f ns\n", IncomingTime);
	
	double delay = MeziramcovaMezera + negexp(PortRamce[PortIndex]);
	TRequest *rq = (new TRequest);
	rq->PortIndex = PortIndex;
	rq->VelikostRamce = randint(64, 32768);
	rq->ActivateDelay((double)rq->VelikostRamce / Speed + delay);               // scheduling incoming of next request
	
	int r = randint(0, 99); // 0..99%
	int OutPort = 0;
	int a = 0;
	while (true) {
		a += PortTable[PortIndex][OutPort];
		if (r < a) break;
		OutPort++;
	}

	CHead *Q = Queue[OutPort];

	if (Q->Empty()) {                            // if the queue is empty
		Into(Q);                                  // inserting request into queue
		printf("Probouzim mikroprocesor %d....\n", OutPort + 1);             
		Service[OutPort]->ActivateAt(Time ());                  // waking up service process
  } else {
		printf("Je tu %d ramcu ve fronte, pridavam dalsi ...\n",Q->Cardinal());
		Into(Q);                                  // ranking the request process to the end of the queue
  }
  
	QBytes[OutPort] += VelikostRamce;
	PocetRamcu++;
	PocetRamcuP[OutPort]++;
	PocetRamcuP2[OutPort]++;

	// Stat
  
  if (PocetRamcuP2[OutPort] > MaxDelkaFront[OutPort]) {
		MaxDelkaFront[OutPort] = PocetRamcuP2[OutPort];
  }
  if (QBytes[OutPort] > MaxDelkaFrontB[OutPort]) {
		MaxDelkaFrontB[OutPort] = QBytes[OutPort];
  }

	
	printf("Cekam na obslouzeni mikroprocesorem ... \n");
	Passivate();                                      // waiting for the finishing the service
	
  PocetRamcuP2[OutPort]--;
	QBytes[OutPort] -= VelikostRamce;
	
	double OutTime = Time();
	printf("Ramec vyslan v case %f ns\n", OutTime);	
	// Stat
	OutTime = OutTime - IncomingTime;
	if (OutTime > MaximalniDoba) MaximalniDoba = OutTime;	
	CelkovaDoba += OutTime;

	if (OutTime > MaximalniDobaP[OutPort]) MaximalniDobaP[OutPort] = OutTime;	
	CelkovaDobaP[OutPort] += OutTime;
	
}

void TService::Run()                               
{
	TRequest* P;                                       // temp. variable
  while(true) {
		CHead *Q = Queue[Id];
			if (Q->Empty()) {
				printf("Fronta %d je prazdna, mikroprocesor je neaktivni ... \n", Id + 1);
				Passivate();                                 // service is waiting for incoming requests
			} else {
				printf("Mikrorocesor %d je probuzen v case %f ns\n", Id, Time ());     // in the queue, there is at least one request
				P=(TRequest*)Q->First();
				P->Out();                                      // removing the request from the queue
				// Posilani ramce
				double t = (double)P->VelikostRamce / Speed + 0; // + zpracovani ramce
				printf("Odesilam ramec velikosti %d po dobu %f ns\n", P->VelikostRamce, t);
				Hold(t);                             // servicing time
				P->ActivateAt (Time());                        // executing request
				printf("Ramec cely odeslan z portu %d v case %f ns\n", Id, Time ());
				Hold(MeziramcovaMezera);
			}
	}
}
/*----------------------------------------------*/
void CSimulation::Run()                          // main process
{
	printf("--- Demonstration of SIMULATION module -- time %f\n", Time());

	srand((unsigned)time( NULL ) );               // inicialization on random number generator
	int i;
	for (i = 0; i < PortCount; i++) {
		QBytes[i] = 0;
		MaxDelkaFront[i] = 0;
		Queue[i]  = new CHead();                          // creating instance of queue
		Service[i] = new TService();                      // creating service process
		Service[i] ->Id = i;
	
		double delay = MeziramcovaMezera + negexp(PortRamce[i]);
		TRequest *rq = (new TRequest);
		rq->PortIndex = i;
		rq->VelikostRamce = randint(64, 32768);
		rq->ActivateDelay((double)rq->VelikostRamce / Speed + delay);             

	}

	Hold(1000000000); // ns, waiting before simulation end 

	printf("***\n");
	printf("Celkova doba: %f, max %f\n", (double)CelkovaDoba / PocetRamcu, MaximalniDoba);

	for (i = 0; i < PortCount; i++) {
		printf("Doba portu %d: %f, max %f\n", i + 1, (double)CelkovaDobaP[i] / PocetRamcuP[i], MaximalniDobaP[i]);
		printf("Maximalni delka fronty %d je %d bajtu, %d ramcu\n", i + 1, MaxDelkaFrontB[i], MaxDelkaFront[i]);
		printf("Portem %d proslo %d, uvizlo %d ramcu\n", i + 1, PocetRamcuP[i], Queue[i] -> Cardinal());
	}

	printf("End of simulation --- time: %f\n\n", Time ());
	int pause; cin >> pause;
	exit(0);
};
